######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################

# Important note:
#  To detect the installation environment, this module makes use of the path in
#  `__file__`. If you move to another place make sure you properly update any line using
#  Path(__file__).

"""Classes implementing Khiops Python' backend runners"""

import io
import os
import platform
import shlex
import shutil
import subprocess
import tempfile
import uuid
import warnings
from abc import ABC, abstractmethod
from pathlib import Path

import khiops
import khiops.core.internals.filesystems as fs
from khiops.core.exceptions import KhiopsEnvironmentError, KhiopsRuntimeError
from khiops.core.internals.common import (
    CommandLineOptions,
    SystemSettings,
    invalid_keys_message,
    is_string_like,
    type_error_message,
)
from khiops.core.internals.io import KhiopsOutputWriter
from khiops.core.internals.task import KhiopsTask
from khiops.core.internals.version import KhiopsVersion


def _isdir_without_all_perms(dir_path):
    """Returns True if the path is a directory but missing of any of rwx permissions"""
    return os.path.isdir(dir_path) and not os.access(
        dir_path, os.R_OK | os.W_OK | os.X_OK
    )


def _get_dir_status(a_dir):
    """Returns the status of a local or remote directory

    Against a local directory a real check is performed. A remote directory is detected
    but not checked.
    """
    if fs.is_local_resource(a_dir):
        # Remove initial slash on windows systems
        # urllib's url2pathname does not work properly
        a_dir_res = fs.create_resource(os.path.normpath(a_dir))
        a_dir_path = a_dir_res.uri_info.path
        if platform.system() == "Windows":
            if a_dir_path.startswith("/"):
                a_dir_path = a_dir_path[1:]

        if not os.path.exists(a_dir_path):
            status = "non-existent"
        elif not os.path.isdir(a_dir_path):
            status = "not-a-dir"
        else:
            status = "ok"
    else:
        status = "remote-path"

    assert status in ["non-existent", "not-a-dir", "ok", "remote-path"]
    return status


def _check_samples_dir(samples_dir):
    # Warn if there are problems with the samples_dir
    samples_dir_status = _get_dir_status(samples_dir)
    download_msg = (
        "Execute the kh-download-datasets script or "
        "the khiops.tools.download_datasets function to download them."
    )
    if samples_dir_status == "non-existent":
        warnings.warn(
            "Sample datasets location does not exist "
            f"({samples_dir}). {download_msg}",
            stacklevel=3,
        )
    elif samples_dir_status == "not-a-dir":
        warnings.warn(
            "Sample datasets location is not a directory "
            f"({samples_dir}). {download_msg}",
            stacklevel=3,
        )


def _extract_path_from_uri(uri):
    res = fs.create_resource(uri)
    if platform.system() == "Windows":
        # Case of file:///<LETTER>:/<REST_OF_PATH>:
        #   Eliminate first slash ("/") from path if the first component
        if (
            res.uri_info.scheme == ""
            and res.uri_info.path[0] == "/"
            and res.uri_info.path[1].isalpha()
            and res.uri_info.path[2] == ":"
        ):
            path = res.uri_info.path[1:]
        # Case of C:/<REST_OF_PATH>:
        #   Just use the original path
        elif len(res.uri_info.scheme) == 1:
            path = uri
        # Otherwise return URI path as-is
        else:
            path = res.uri_info.path

    else:
        path = res.uri_info.path
    return path


def _khiops_env_file_exists(env_dir):
    """Check ``khiops_env`` exists relative to the specified environment dir"""
    khiops_env_path = os.path.join(env_dir, "khiops_env")
    if platform.system() == "Windows":
        khiops_env_path += ".cmd"
    return os.path.exists(khiops_env_path) and os.path.isfile(khiops_env_path)


def _infer_env_bin_dir_for_conda_based_installations():
    """Infer reference directory for Conda-based Khiops installations"""
    assert os.path.basename(Path(__file__).parents[2]) == "khiops", (
        f"The {os.path.basename(__file__)} file has been moved. "
        "Please fix the `Path.parents` in this method "
        "so it finds the conda environment directory of this module"
    )

    # Obtain the full path of the current file
    current_file_path = Path(__file__).resolve()

    # Windows: Match %CONDA_PREFIX%\Lib\site-packages\khiops\core\internals\runner.py
    if platform.system() == "Windows":
        conda_env_dir = current_file_path.parents[5]
    # Linux/macOS:
    # Match $CONDA_PREFIX/[Ll]ib/python3.X/site-packages/khiops/core/internals/runner.py
    else:
        conda_env_dir = current_file_path.parents[6]

    # Conda env binary dir is:
    # - on Windows: conda_env_dir\Library\bin
    # - on Linux/macOS: conda_env_dir\bin
    if platform.system() == "Windows":
        env_bin_dir = os.path.join(str(conda_env_dir), "Library", "bin")
    else:
        env_bin_dir = os.path.join(str(conda_env_dir), "bin")

    return env_bin_dir


def _check_conda_env_bin_dir(conda_env_bin_dir):
    """Check inferred Conda environment binary directory really is one

    A real Conda environment binary directory:
    - should exist
    - should not be directly under the root directory
    - should coexist with `conda-meta` directory under the same parent
    """
    conda_env_bin_dir_path = Path(conda_env_bin_dir)

    # Conda env bin dir should end with `/bin`
    assert conda_env_bin_dir_path.parts[-1] == "bin"

    is_conda_env_bin_dir = False

    # Conda env dir is not equal to its root dir
    # Conda env bin dir exists, along with the `conda-meta` dir
    # Note: On Windows, Conda env bin dir equals conda env dir\Library\bin
    conda_env_dir_path = conda_env_bin_dir_path.parent
    if platform.system() == "Windows":
        conda_env_dir_path = conda_env_dir_path.parent
    if (
        str(conda_env_dir_path) != conda_env_dir_path.root  # `.root` is an `str`
        and conda_env_bin_dir_path.is_dir()
        and conda_env_dir_path.joinpath("conda-meta").is_dir()
    ):
        is_conda_env_bin_dir = True
    return is_conda_env_bin_dir


def _infer_khiops_installation_method(trace=False):
    """Return the Khiops installation method"""
    # We are in a Conda environment if
    # - the CONDA_PREFIX environment variable exists and,
    # - the khiops_env script exists within:
    #   - `%CONDA_PREFIX\Library\bin%` on Windows
    #   - `$CONDA_PREFIX/bin` on Linux and MacOS
    # Note: The check that the Khiops binaries are actually executable is done
    #       afterwards by the initializations method.
    installation_method = "unknown"
    if "CONDA_PREFIX" in os.environ:
        conda_env_dir = os.environ["CONDA_PREFIX"]
        if platform.system() == "Windows":
            conda_binary_dir = os.path.join(conda_env_dir, "Library", "bin")
        else:
            conda_binary_dir = os.path.join(conda_env_dir, "bin")
        if _khiops_env_file_exists(conda_binary_dir):
            installation_method = "conda"
    # Otherwise (installation_method is still "unknown"), we choose between
    # conda-based and local (default choice)
    if installation_method == "unknown":
        env_bin_dir = _infer_env_bin_dir_for_conda_based_installations()
        if trace:
            print(f"Environment binary dir: '{env_bin_dir}'")
        if _check_conda_env_bin_dir(env_bin_dir) and _khiops_env_file_exists(
            env_bin_dir
        ):
            installation_method = "conda-based"
        else:
            installation_method = "binary+pip"
    if trace:
        print(f"Installation method: '{installation_method}'")
    assert installation_method in ("conda", "conda-based", "binary+pip")
    return installation_method


def _check_executable(bin_path):
    if not os.path.isfile(bin_path):
        raise KhiopsEnvironmentError(f"Non-regular executable file. Path: {bin_path}")
    elif not os.access(bin_path, os.X_OK):
        raise KhiopsEnvironmentError(
            f"Executable has no execution rights. Path: {bin_path}"
        )


class KhiopsRunner(ABC):
    """Abstract Khiops Python runner to be re-implemented"""

    def __init__(self):
        """See class docstring"""
        self._initialize_root_temp_dir()
        self._khiops_version = None
        self._samples_dir = None

        # Whether to write the Khiops Python library version of the scenarios
        # For development uses only
        self._write_version = True

    def _initialize_root_temp_dir(self):
        """Initializes the runner's root temporary directory

        It tries to set a proper root temporary directory. It tries the following
        strategies in order:
        - Check that ``$TEMP/khiops/python`` exists and use it
        - Try to create ``$TEMP/khiops/python` and use it
        - Create a ``$TEMP/khiops_<HASH>/python`` and use it
        """

        # Create the directory if it doesn't exists
        self._root_temp_dir = os.path.join(tempfile.gettempdir(), "khiops", "python")
        if not os.path.exists(self._root_temp_dir):
            os.makedirs(self._root_temp_dir)
        # Create the dir with a hash name if it is a dir but it doesn't have all
        # permissions or if it is a file
        elif os.path.isfile(self._root_temp_dir) or _isdir_without_all_perms(
            self._root_temp_dir
        ):
            self._root_temp_dir = os.path.join(
                tempfile.mkdtemp(prefix="khiops_"), "python"
            )
            os.makedirs(self._root_temp_dir, exist_ok=True)

    @property
    def root_temp_dir(self):
        r"""str: The runner's temporary directory

        The temporary scenarios/templates and dictionary files created by
        the Khiops Python library are stored here.

        Default value:
            - Windows: ``%TEMP%\khiops\python``
            - Linux: ``$TMP/khiops/python``

        When set to a local path it tries to create the specified directory if it
        doesn't exist.

        Raises
        ------
        `.KhiopsEnvironmentError`
            If set to a local path: if it is a file or if it does not have ``+rwx``
            permissions.
        """
        return self._root_temp_dir

    @root_temp_dir.setter
    def root_temp_dir(self, dir_path):
        # Check existence, directory status and permissions for local paths
        if fs.is_local_resource(dir_path):
            real_dir_path = _extract_path_from_uri(dir_path)
            if os.path.exists(real_dir_path):
                if os.path.isfile(real_dir_path):
                    raise KhiopsEnvironmentError(
                        f"File at temporary directory os.path. Path: {real_dir_path}"
                    )
                elif _isdir_without_all_perms(real_dir_path):
                    raise KhiopsEnvironmentError(
                        "Temporary directory must have +rwx permissions. "
                        f"Path: {real_dir_path}"
                    )
            else:
                os.makedirs(real_dir_path)
        # There are no checks for non local filesystems (no `else` statement)
        self._root_temp_dir = dir_path

    def create_temp_file(self, prefix, suffix):
        """Creates a unique temporary file in the runner's root temporary directory

        .. note::
            For remote filesystems no actual file is created, just a (highly probable)
            unique path is returned.

        Parameters
        ----------
        prefix : str
            Prefix for the file's name.

        suffix : str
            Suffix for the file's name.

        Returns
        -------
        str
            A unique path within the root temporary directory. The file is created only
            in the case of a local filesystem.
        """
        # Local resource: Effectively create the file with the python file API
        if fs.is_local_resource(self.root_temp_dir):
            # Extract the path from the potential URI
            root_temp_dir_path = _extract_path_from_uri(self.root_temp_dir)

            # Create the temporary file
            tmp_file_fd, tmp_file_path = tempfile.mkstemp(
                prefix=prefix, suffix=suffix, dir=root_temp_dir_path
            )
            os.close(tmp_file_fd)
        # Remote resource: Just return a highly probable unique path
        else:
            tmp_file_path = fs.get_child_path(
                self.root_temp_dir, f"{prefix}{uuid.uuid4()}{suffix}"
            )

        return tmp_file_path

    def create_temp_dir(self, prefix):
        """Creates a unique directory in the runner's root temporary directory

        Parameters
        ----------
        prefix : str
            Prefix for the directory's name.

        Returns
        -------
        str
            A unique directory path within the root temporary directory. The directory
            is created only in the case of a local filesystem.
        """
        # Local resource: Effectively create the directory with the python file API
        if fs.is_local_resource(self.root_temp_dir):
            root_temp_dir_path = _extract_path_from_uri(self.root_temp_dir)
            temp_dir = tempfile.mkdtemp(prefix=prefix, dir=root_temp_dir_path)
        # Remote resource: Just return a highly probable unique path
        else:
            temp_dir = fs.get_child_path(self.root_temp_dir, f"{prefix}{uuid.uuid4()}")
        return temp_dir

    @property
    def samples_dir(self):
        r"""str: Location of the Khiops' sample datasets directory. May be an URL/URI"""
        return self._get_samples_dir()

    def _get_samples_dir(self):
        """To be overridden by subclasses"""
        return self._samples_dir

    @samples_dir.setter
    def samples_dir(self, samples_dir):
        if not is_string_like(samples_dir):
            raise TypeError(
                type_error_message("samples_dir", samples_dir, "string-like")
            )
        self._set_samples_dir(samples_dir)

    def _set_samples_dir(self, samples_dir):
        """To be overridden by child classes to add additional checks"""
        self._samples_dir = samples_dir

    @property
    def khiops_version(self):
        """`.KhiopsVersion`: The version of the Khiops backend of this runner"""
        return self._get_khiops_version()

    def _get_khiops_version(self):
        """khiops_version getter to be overridden by subclasses"""
        return self._khiops_version

    def _build_status_message(self):
        """Constructs the status message

        Descendant classes can add additional information.

        Returns
        -------
        tuple
            A 2-tuple containing:
            - The status message
            - A list of warning messages
        """
        # Capture the status of the the samples dir
        warning_list = []
        with warnings.catch_warnings(record=True) as caught_warnings:
            samples_dir_path = self.samples_dir
        if caught_warnings is not None:
            warning_list += caught_warnings

        status_msg = "Khiops Python library settings\n"
        status_msg += f"version             : {khiops.__version__}\n"
        status_msg += f"runner class        : {self.__class__.__name__}\n"
        status_msg += f"root temp dir       : {self.root_temp_dir}\n"
        status_msg += f"sample datasets dir : {samples_dir_path}\n"
        status_msg += f"package dir         : {Path(__file__).parents[2]}\n"
        return status_msg, warning_list

    def print_status(self):
        """Prints the status of the runner to stdout"""
        # Obtain the status_msg, errors and warnings
        try:
            status_msg, warning_list = self._build_status_message()
        except (KhiopsEnvironmentError, KhiopsRuntimeError) as error:
            print(f"Khiops Python library status KO: {error}")
            return 1

        # Print status details
        print(status_msg, end="")

        # Print status
        print("Khiops Python library status OK", end="")
        if warning_list:
            print(", with warnings:")
            for warning in warning_list:
                print(f"warning: {warning.message}")
        else:
            print("")
        return 0

    @abstractmethod
    def _initialize_khiops_version(self):
        """Initialization of `khiops_version` to be implemented in child classes"""

    def run(
        self,
        task,
        task_args,
        command_line_options=None,
        trace=False,
        system_settings=None,
        stdout_file_path="",
        stderr_file_path="",
        force_ansi_scenario=False,
        **kwargs,
    ):
        """Runs a Khiops Task

        Parameters
        ----------
        task : `.KhiopsTask`
            Khiops task to be run.
        task_args : dict
            Arguments for the task.
        command_line_options : `.CommandLineOptions`, optional
            Command line options for all tasks. If not set the default values are used.
            See the `.CommandLineOptions` for more information.
        trace : bool, default ``False``
            If True prints the command line executed of the process and does not delete
            any temporary files created.
        system_settings : `.SystemSettings`, optional
            *Advanced:* System settings for all tasks. See the `.SystemSettings`
            class for more information.
        stdout_file_path : str, default ""
            *Advanced* Path to a file where the Khiops process writes its stdout stream.
            Normally Khiops should not write to this stream but MPI, filesystems plugins
            or debug versions may do it. The stream is captured with a UTF-8 encoding
            and replacing encoding errors. If equal to "" then it writes no file.
        stderr_file_path : str, default ""
            *Advanced* Path to a file where the Khiops process writes its stderr stream.
            Normally Khiops should not write to this stream but MPI, filesystems plugins
            or debug versions may do it. The stream is captured with a UTF-8 encoding
            and replacing encoding errors. If equal to "" then it writes no file.
        force_ansi_scenario : bool, default ``False``
            *Advanced:* If True the internal scenario generated by Khiops will force
            characters such as accentuated ones to be decoded with the UTF8->ANSI khiops
            transformation.

        Raises
        ------
        `ValueError`
            - Unknown keyword argument
            - Files or executable not found
            - Errors in the execution of the Khiops tool

        `TypeError`
            - Invalid type of a keyword argument
            - When the search/replace pairs are not strings
        """
        # Warn if there are still kwargs: At this point any keyword argument is invalid
        if kwargs:
            warnings.warn(invalid_keys_message(kwargs), stacklevel=3)
            kwargs.clear()

        # Use the default command line options if not specified
        if command_line_options is None:
            command_line_options = CommandLineOptions()

        # Check the call arguments
        if not isinstance(trace, bool):
            raise TypeError(type_error_message("trace", trace, bool))
        command_line_options.check()
        if not isinstance(stdout_file_path, str):
            raise TypeError(
                type_error_message("stdout_file_path", stdout_file_path, str)
            )
        if not isinstance(stderr_file_path, str):
            raise TypeError(
                type_error_message("stderr_file_path", stderr_file_path, str)
            )

        # Write the scenarios file
        scenario_path = self._write_task_scenario_file(
            task,
            task_args,
            system_settings,
            force_ansi_scenario=force_ansi_scenario,
        )

        # If no log file specified: Use a temporary file
        tmp_log_file_path = None
        if not command_line_options.log_file_path:
            tmp_log_file_path = self.create_temp_file("_run_", ".log")
            command_line_options.log_file_path = tmp_log_file_path

        # Execute Khiops
        try:
            # Disable pylint warning about abstract method _run returning None
            # pylint: disable=assignment-from-no-return
            return_code, stdout, stderr = self._run(
                task.tool_name,
                scenario_path,
                command_line_options,
                trace,
            )
            # pylint: enable=assignment-from-no-return
        # Catch an OS level error if any
        except OSError as error:
            raise KhiopsRuntimeError("Khiops execution failed.") from error
        # Report any errors raised by Khiops
        else:
            # Write the stdout and stderr streams if specified
            if stdout_file_path:
                fs.write(stdout_file_path, bytes(stdout, encoding="utf8"))
            if stderr_file_path:
                fs.write(stderr_file_path, bytes(stderr, encoding="utf8"))

            # Report the exit status
            self._report_exit_status(
                task.tool_name,
                return_code,
                stdout,
                stderr,
                command_line_options.log_file_path,
            )
        # Clean files unless trace mode is activated
        finally:
            if trace:
                print(f"Khiops execution scenario: {scenario_path}")
                print(f"Khiops log file: {command_line_options.log_file_path}")
            else:
                fs.remove(scenario_path)
                if tmp_log_file_path is not None:
                    fs.remove(tmp_log_file_path)

    def _report_exit_status(
        self, tool_name, return_code, stdout, stderr, log_file_path
    ):
        """Reports the exit status of a Khiops execution"""
        # Note:
        #   We report stdout and stderr below because we use a log file and thus
        #   normally Khiops doesn't write anything to these streams. In practice MPI and
        #   the remote filesystems plugins may write to them to report anomalies.

        # Report messages:
        # - The warnings in the log
        # - The errors and/or fatal errors in the log
        # - The stdout if not empty
        # - The stderr if not empty
        #
        # If there were any errors (fatal or not) or the return code is non-zero the
        # reporting is via an exception. Otherwise we show the message as a warning.
        #

        # Create the message reporting the errors and warnings
        error_msg = ""
        # If the log file exists: Collect the errors and warnings messages
        if fs.exists(log_file_path):
            errors, fatal_errors, warning_messages = self._collect_errors(log_file_path)
            if warning_messages:
                error_msg += "Warnings in log:\n" + "".join(warning_messages)
            if errors:
                if error_msg:
                    error_msg += "\n"
                error_msg += "Errors in log:\n" + "".join(errors)
            if fatal_errors:
                if error_msg:
                    error_msg += "\n"
                error_msg += "Fatal errors in log:\n" + "".join(fatal_errors)
        # Otherwise warn that the log file is missing
        else:
            warnings.warn(
                f"Log file not found after {tool_name} execution."
                f"Path: {log_file_path}"
            )
            errors = fatal_errors = []

        # Add stdout to the warning message if non empty
        if stdout:
            if error_msg:
                error_msg += "\n"
            error_msg += f"Contents of stdout:\n{stdout}"

        # Add stderr to the warning message if non empty
        if stderr:
            if error_msg:
                error_msg += "\n"
            error_msg += f"Contents of stderr:\n{stderr}"

        # Report the message to the user if there were any
        if error_msg:
            # Raise an exception if there were errors
            if errors or fatal_errors or return_code != 0:
                raise KhiopsRuntimeError(
                    f"{tool_name} execution had errors (return code {return_code}):\n"
                    f"{error_msg}"
                )
            # Otherwise show the message as a warning
            else:
                error_msg = (
                    f"Khiops ended correctly but there were minor issues:\n{error_msg}"
                )
                warnings.warn(error_msg.rstrip())
        # Raise an exception anyway for a non-zero return code without any message
        else:
            if return_code != 0:
                raise KhiopsRuntimeError(
                    f"{tool_name} execution had errors (return code {return_code}) "
                    "but no message is available"
                )

    def _collect_errors(self, log_file_path):
        # Collect errors any errors found in the log
        errors = []
        fatal_errors = []
        warning_messages = []

        # Look in the log for error lines
        log_file_lines = None
        try:
            log_file_contents = fs.read(log_file_path)
            log_file_lines = io.TextIOWrapper(
                io.BytesIO(log_file_contents), encoding="utf8", errors="replace"
            )
            for line_number, line in enumerate(log_file_lines, start=1):
                if line.startswith("warning : "):
                    warning_messages.append(f"Line {line_number}: {line}")
                elif line.startswith("error : "):
                    errors.append(f"Line {line_number}: {line}")
                elif line.startswith("fatal error : "):
                    fatal_errors.append(f"Line {line_number}: {line}")

        # Warn on error for remote file handling. Replace with empty log file.
        except ImportError:
            warnings.warn(
                "Could not read remote log file and errors may not be "
                "reported. Make sure you have installed the extra "
                "dependencies for remote filesystems.",
                stacklevel=3,
            )

        return errors, fatal_errors, warning_messages

    def _create_scenario_file(self, task):
        assert isinstance(task, KhiopsTask)
        return self.create_temp_file(f"{task.name}_", "._kh")

    def _write_task_scenario_file(
        self, task, task_args, system_settings, force_ansi_scenario=False
    ):
        scenario_path = self._create_scenario_file(task)
        with io.BytesIO() as scenario_stream:
            writer = KhiopsOutputWriter(scenario_stream, force_ansi=force_ansi_scenario)
            if self._write_version:
                writer.writeln(f"// Generated by khiops-python {khiops.__version__}")
            self._write_task_scenario(writer, task, task_args, system_settings)
            fs.write(scenario_path, scenario_stream.getvalue())

        return scenario_path

    def _write_task_scenario(self, writer, task, task_args, system_settings):
        assert isinstance(task, KhiopsTask)
        assert isinstance(task_args, dict)
        assert isinstance(system_settings, SystemSettings)

        # Write the task scenario
        self._write_scenario_prologue(writer, system_settings)
        task.write_execution_scenario(writer, task_args)
        self._write_scenario_exit_statement(writer)

    def _write_scenario_prologue(self, writer, system_settings):
        # Write the system settings if any
        if (
            system_settings.max_cores
            or system_settings.memory_limit_mb
            or system_settings.temp_dir
        ):
            writer.writeln("// System settings")
            if system_settings.max_cores:
                writer.write("AnalysisSpec.SystemParameters.MaxCoreNumber ")
                writer.writeln(str(system_settings.max_cores))
            if system_settings.memory_limit_mb:
                writer.write("AnalysisSpec.SystemParameters.MemoryLimit ")
                writer.writeln(str(system_settings.memory_limit_mb))
            if system_settings.temp_dir:
                writer.write("AnalysisSpec.SystemParameters.TemporaryDirectoryName ")
                writer.writeln(system_settings.temp_dir)
            writer.writeln("")

        # Write the user defined prologue
        if system_settings.scenario_prologue:
            writer.writeln("// User-defined prologue")
            for line in system_settings.scenario_prologue.split("\n"):
                writer.writeln(line)
            writer.writeln("")

    def _write_scenario_exit_statement(self, writer):
        writer.writeln("")
        writer.writeln("// Exit Khiops")
        writer.writeln("ClassManagement.Quit")
        writer.writeln("OK")

    @abstractmethod
    def _run(
        self,
        tool_name,
        scenario_path,
        command_line_options,
        trace,
    ):
        """Abstract run method to be implemented in child classes

        Returns
        -------
        tuple
            A 3-tuple containing the return code, the stdout and the stderr of the
            Khiops process.

        Raises
        ------
        `.KhiopsRuntimeError`
            If there were any errors in the Khiops execution.
        """


class KhiopsLocalRunner(KhiopsRunner):
    r"""Implementation of a local Khiops runner

    Requires either:

    - This package installed through Conda and run from a Conda environment, or
    - the ``khiops-core`` Linux native package installed on the local machine, or
    - the Windows Khiops desktop application installed on the local machine

    .. rubric:: Samples directory settings

    Default values for the ``samples_dir`` attribute:

    - The value of the ``KHIOPS_SAMPLES_DIR`` environment variable (path to the Khiops
      sample datasets directory).
    - Otherwise:

      - Windows:

        - ``%PUBLIC%\khiops_data\samples%`` if it exists and is a directory
        - ``%USERPROFILE%\khiops_data\samples%`` otherwise

      - Linux and macOS:

        - ``$HOME/khiops_data/samples``

    """

    def __init__(self):
        # Define specific attributes
        self._mpi_command_args = None
        self._khiops_path = None
        self._khiops_coclustering_path = None
        self._khiops_version = None
        self._samples_dir = None
        self._samples_dir_checked = False

        # Call parent constructor
        super().__init__()

        # Initialize Khiops environment
        self._initialize_khiops_environment()

    def _initialize_khiops_environment(self):
        # Check the `khiops_env` script
        # On Windows native installations, rely on the `KHIOPS_HOME` environment
        # variable set by the Khiops Desktop Application installer
        installation_method = _infer_khiops_installation_method()
        if platform.system() == "Windows" and installation_method == "binary+pip":
            # KHIOPS_HOME variable by default
            if "KHIOPS_HOME" in os.environ:
                khiops_env_path = os.path.join(
                    os.environ["KHIOPS_HOME"], "bin", "khiops_env.cmd"
                )
            # Raise error if KHIOPS_HOME is not set
            else:
                raise KhiopsEnvironmentError(
                    "No environment variable named 'KHIOPS_HOME' found. "
                    "Make sure you have installed Khiops >= 10.2.3. "
                    "Go to https://khiops.org for more information."
                )

        # In Conda-based environments, `khiops_env` might not be in the PATH,
        # hence its path must be inferred
        elif installation_method == "conda-based":
            khiops_env_path = os.path.join(
                _infer_env_bin_dir_for_conda_based_installations(), "khiops_env"
            )
            if platform.system() == "Windows":
                khiops_env_path += ".cmd"

        # On UNIX or Conda, khiops_env is always in path for a proper installation
        else:
            khiops_env_path = shutil.which("khiops_env")
            if khiops_env_path is None:
                raise KhiopsEnvironmentError(
                    "The 'khiops_env' script not found for the current "
                    f"'{installation_method}' installation method. Make sure "
                    "you have installed khiops >= 10.2.3. "
                    "Go to https://khiops.org for more information."
                )

        with subprocess.Popen(
            [khiops_env_path, "--env"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        ) as khiops_env_process:
            stdout, stderr = khiops_env_process.communicate()
            if khiops_env_process.returncode != 0:
                raise KhiopsEnvironmentError(
                    "Error initializing the environment for Khiops from the "
                    f"{khiops_env_path} script. Contents of stderr:\n{stderr}"
                )
            for line in stdout.split("\n"):
                tokens = line.rstrip().split(maxsplit=1)
                if len(tokens) == 2:
                    var_name, var_value = tokens
                elif len(tokens) == 1:
                    var_name = tokens[0]
                    var_value = ""
                else:
                    continue
                # Set paths to Khiops binaries
                if var_name == "KHIOPS_PATH":
                    self.khiops_path = var_value
                    os.environ["KHIOPS_PATH"] = var_value
                elif var_name == "KHIOPS_COCLUSTERING_PATH":
                    self.khiops_coclustering_path = var_value
                    os.environ["KHIOPS_COCLUSTERING_PATH"] = var_value
                # Set MPI command
                elif var_name == "KHIOPS_MPI_COMMAND":
                    self._mpi_command_args = shlex.split(var_value)
                    os.environ["KHIOPS_MPI_COMMAND"] = var_value
                # Propagate all the other environment variables to Khiops binaries
                else:
                    os.environ[var_name] = var_value

        # Check the tools exist and are executable
        self._check_tools()

        # Initialize the default samples dir
        self._initialize_default_samples_dir()

    def _initialize_default_samples_dir(self):
        """See class docstring"""
        # Set the fallback value for the samples directory
        home_samples_dir = Path.home() / "khiops_data" / "samples"

        # Take the value of an environment variable in priority
        if "KHIOPS_SAMPLES_DIR" in os.environ:
            self._samples_dir = os.environ["KHIOPS_SAMPLES_DIR"]

        # The samples location of Windows systems is:
        # - %PUBLIC%\khiops_data\samples if %PUBLIC% exists
        # - %USERPROFILE%\khiops_data\samples otherwise
        elif platform.system() == "Windows":
            if "PUBLIC" in os.environ:
                public_samples_dir = os.path.join(
                    os.environ["PUBLIC"], "khiops_data", "samples"
                )
            else:
                public_samples_dir = None

            ok_statuses = ["ok", "remote"]
            if (
                public_samples_dir is not None
                and _get_dir_status(public_samples_dir) in ok_statuses
            ):
                self._samples_dir = public_samples_dir
            else:
                self._samples_dir = str(home_samples_dir)

        # The default samples location on Unix systems is:
        # $HOME/khiops/samples on Linux and Mac OS
        else:
            self._samples_dir = str(home_samples_dir)

        assert self._samples_dir is not None

    def _check_tools(self):
        """Checks the that the tool binaries exist and are executable"""
        for tool_name in ["khiops", "khiops_coclustering"]:
            _check_executable(self._tool_path(tool_name))

    def _initialize_khiops_version(self):
        # Run khiops with the -v
        stdout, stderr, return_code = self.raw_run("khiops", ["-v"], use_mpi=False)

        # On success parse and save the version
        if return_code == 0:
            # Skip potential non-version lines
            # (ex: Completed loading of file driver, debug info ...)
            for line in stdout.split("\n"):
                line = line.strip("\r")  # remove Windows-specific Carriage-Return
                if line.startswith("Khiops"):
                    khiops_version_str = line.rstrip().split(" ")[1]
                    break
        # If MODL -v fails we raise an informative error
        else:
            error_msg = (
                f"Could not obtain the Khiops version."
                f"\nError executing '{self._tool_path('khiops')} -v': "
                f"return code {return_code}."
            )
            error_msg += f"\nstdout: {stdout}" if stdout else ""
            error_msg += f"\nstderr: {stderr}" if stderr else ""
            raise KhiopsRuntimeError(error_msg)

        self._khiops_version = KhiopsVersion(khiops_version_str)

        # Warn if the khiops version is too far from the Khiops Python library version
        compatible_khiops_version = khiops.get_compatible_khiops_version()
        if self._khiops_version.major > compatible_khiops_version.major:
            warnings.warn(
                f"Khiops version '{self._khiops_version}' is ahead of "
                f"the Khiops Python library version '{khiops.__version__}'. "
                "There may be compatibility errors and "
                "we recommend you to update to the latest Khiops Python "
                "library version. See https://khiops.org for more information.",
                stacklevel=3,
            )

    def _build_status_message(self):
        # Call the parent's method
        status_msg, warning_list = super()._build_status_message()

        # Build the messages for install type and mpi
        install_type_msg = _infer_khiops_installation_method()
        if self._mpi_command_args:
            mpi_command_args_msg = " ".join(self._mpi_command_args)
        else:
            mpi_command_args_msg = "<empty>"

        # Build the message
        status_msg += "\n\n"
        status_msg += "khiops local installation settings\n"
        status_msg += f"version             : {self.khiops_version}\n"
        status_msg += f"Khiops path         : {self.khiops_path}\n"
        status_msg += f"Khiops CC path      : {self.khiops_coclustering_path}\n"
        status_msg += f"install type        : {install_type_msg}\n"
        status_msg += f"MPI command         : {mpi_command_args_msg}\n"

        # Add output of khiops -s which gives the MODL_* binary status
        status_msg += "\n\n"
        khiops_executable = os.path.join(os.path.dirname(self.khiops_path), "khiops")
        status_msg += f"Khiops executable status (output of '{khiops_executable} -s')\n"
        stdout, stderr, return_code = self.raw_run("khiops", ["-s"], use_mpi=True)

        # On success retrieve the status and added to the message
        if return_code == 0:
            status_msg += stdout
        else:
            warning_list.append(stderr)
        status_msg += "\n"

        return status_msg, warning_list

    def _get_khiops_version(self):
        # Initialize the first time it is called
        if self._khiops_version is None:
            self._initialize_khiops_version()
        assert isinstance(self._khiops_version, KhiopsVersion), type_error_message(
            self._khiops_version, "khiops_version", KhiopsVersion
        )
        return self._khiops_version

    @property
    def mpi_command_args(self):
        return self._mpi_command_args

    @property
    def khiops_path(self):
        """str: Path to the ``MODL*`` Khiops binary

        Set by the ``khiops_env`` script from the ``khiops-core`` package.

        """
        return self._khiops_path

    @khiops_path.setter
    def khiops_path(self, modl_path):
        # Check that the path is a directory and it exists
        if not os.path.exists(modl_path):
            raise KhiopsEnvironmentError(f"Inexistent Khiops path: '{modl_path}'")
        if not os.path.isfile(modl_path):
            raise KhiopsEnvironmentError(
                f"Khiops file path is a directory: {modl_path}"
            )

        # Set the MODL path
        self._khiops_path = modl_path

    @property
    def khiops_coclustering_path(self):
        """str: Path to the ``MODL_Coclustering`` Khiops Coclustering binary

        Set by the ``khiops_env`` script from the ``khiops-core`` package.

        """
        return self._khiops_coclustering_path

    @khiops_coclustering_path.setter
    def khiops_coclustering_path(self, modl_coclustering_path):
        # Check that the path is a directory and it exists
        if not os.path.exists(modl_coclustering_path):
            raise KhiopsEnvironmentError(
                f"Inexistent Khiops coclustering path: '{modl_coclustering_path}'"
            )
        if not os.path.isfile(modl_coclustering_path):
            raise KhiopsEnvironmentError(
                "Khiops coclustering file path is a directory: "
                f"{modl_coclustering_path}"
            )

        # Set the MODL_Coclustering path
        self._khiops_coclustering_path = modl_coclustering_path

    def _tool_path(self, tool_name):
        """Full path of a Khiops tool binary"""
        assert (
            self.khiops_path is not None and self.khiops_coclustering_path is not None
        )
        tool_name = tool_name.lower()
        if tool_name not in ["khiops", "khiops_coclustering"]:
            raise ValueError(f"Invalid tool name: {tool_name}")
        modl_binaries = {
            "khiops": self.khiops_path,
            "khiops_coclustering": self.khiops_coclustering_path,
        }
        bin_path = modl_binaries[tool_name]

        return bin_path

    def _set_samples_dir(self, samples_dir):
        """Checks and sets the samples directory"""
        _check_samples_dir(samples_dir)
        super()._set_samples_dir(samples_dir)

    def _get_samples_dir(self):
        # Check the samples dir once (the check emits only warnings)
        if not self._samples_dir_checked:
            _check_samples_dir(self._samples_dir)
            self._samples_dir_checked = True
        return self._samples_dir

    def raw_run(self, tool_name, command_line_args=None, use_mpi=True, trace=False):
        """Execute a Khiops tool with given command line arguments

        Parameters
        ----------
        tool_name : {"khiops", "khiops_coclustering"}
            Name of the tool to execute.
        command_line_args : list of str, optional
            Command line arguments of Khiops.
        use_mpi : bool, optional
            Whether to execute the application with MPI
        trace : bool, default False
            If ``True`` print the trace of the process.

        Examples
        --------
        >>> raw_run("khiops", ["-b", "-i" , "scenario._kh"])

        is equivalent to execute in a shell::

            > khiops -b -i scenario._kh
        """
        # Check command_line_args type
        if command_line_args and not isinstance(command_line_args, list):
            raise TypeError(
                type_error_message("command_line_args", command_line_args, list)
            )

        # Build command line arguments
        # Nota: Khiops Coclustering is executed without MPI
        khiops_process_args = []
        if tool_name == "khiops" and use_mpi:
            khiops_process_args += self._mpi_command_args
        khiops_process_args += [self._tool_path(tool_name)]
        if command_line_args:
            khiops_process_args += command_line_args

        # If trace is on: display call arguments
        if trace:
            quote = '"' if platform.system() == "Windows" else "'"
            khiops_call = khiops_process_args[0]
            for arg in khiops_process_args[1:]:
                if isinstance(arg, bytes):
                    arg = arg.decode("utf8", errors="replace")
                if arg.startswith("-"):
                    khiops_call += f" {arg}"
                else:
                    khiops_call += f" {quote}{arg}{quote}"
            print(f"Khiops execution call: {khiops_call}")

        # Execute the process
        with subprocess.Popen(
            khiops_process_args,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf8",
            errors="replace",
            universal_newlines=True,
        ) as khiops_process:
            stdout, stderr = khiops_process.communicate()

        return stdout, stderr, khiops_process.returncode

    def _run(
        self,
        tool_name,
        scenario_path,
        command_line_options,
        trace,
    ):
        # Execute the tool
        khiops_args = command_line_options.build_command_line_options(scenario_path)
        stdout, stderr, return_code = self.raw_run(
            tool_name, command_line_args=khiops_args, trace=trace
        )

        return return_code, stdout, stderr


#########################
# Current Runner Access #
#########################

# Disable pylint UPPER_CASE convention: _khiops_runner is non-constant
# pylint: disable=invalid-name

# Runner (backend) of Khiops Python, by default None for lazy initialization
_khiops_runner = None


def set_runner(runner):
    """Sets the current KhiopsRunner of the module"""
    if not isinstance(runner, KhiopsRunner):
        raise TypeError(type_error_message("runner", runner, KhiopsRunner))
    global _khiops_runner
    _khiops_runner = runner


def get_runner():
    """Returns the current KhiopsRunner of the module

    Returns
    -------
    `.KhiopsRunner`
        The current Khiops Python runner of the module.
    """
    #  Define and initialize a runner for a local Khiops installation
    global _khiops_runner
    if _khiops_runner is None:
        _khiops_runner = KhiopsLocalRunner()
    return _khiops_runner


# pylint: enable=invalid-name



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""extract_keys_from_data_table task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import BoolType, StringLikeType

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "extract_keys_from_data_table",
        "khiops",
        "10.0.0",
        [
            ("dictionary_file_path", StringLikeType),
            ("dictionary_name", StringLikeType),
            ("data_table_path", StringLikeType),
            ("output_data_table_path", StringLikeType),
        ],
        [
            ("detect_format", BoolType, True),
            ("header_line", BoolType, True),
            ("field_separator", StringLikeType, ""),
            ("output_header_line", BoolType, True),
            ("output_field_separator", StringLikeType, ""),
        ],
        [
            "dictionary_file_path",
            "data_table_path",
            "output_data_table_path",
        ],
        # fmt: off
        """
        // Dictionary file
        ClassManagement.OpenFile
        ClassFileName __dictionary_file_path__
        OK

        // Extract keys settings
        LearningTools.ExtractKeysFromDataTable
        ClassName __dictionary_name__
        SourceDataTable.DatabaseName __data_table_path__
        SourceDataTable.HeaderLineUsed __header_line__
        SourceDataTable.FieldSeparator __field_separator__
        __OPT__
        __detect_format__
        SourceDataTable.DatabaseFormatDetector.DetectFileFormat
        __END_OPT__
        TargetDataTable.DatabaseName __output_data_table_path__
        TargetDataTable.HeaderLineUsed __output_header_line__
        TargetDataTable.FieldSeparator __output_field_separator__
        ExtractKeysFromDataTable
        Exit
        """,
        # fmt: on
    ),
]



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Helper functions for specific and/or advanced treatments"""
import os
import platform
import subprocess
import warnings

import khiops.core.internals.filesystems as fs
from khiops.core import api
from khiops.core.dictionary import (
    Dictionary,
    DictionaryDomain,
    Variable,
    read_dictionary_file,
)
from khiops.core.internals.common import (
    create_unambiguous_khiops_path,
    deprecation_message,
    is_list_like,
    type_error_message,
)


def build_multi_table_dictionary_domain(
    dictionary_domain, root_dictionary_name, secondary_table_variable_name
):
    """Builds a multi-table dictionary domain from a dictionary with a key

    .. note::

        This is a special-purpose function whose goal is to assist in preparing the
        coclustering deployment.

        This function builds a new root dictionary and adds it to an existing dictionary
        domain.
        The new root dictionary only contains one field, which references a preexisting
        dictionary from the input dictionary domain as a new (secondary) Table variable.
        The preexisting dictionary must have a key set on it, as this is the join key
        with the new root table.

    .. warning::
        This method is *deprecated* since Khiops 10.3.1.0 and will be removed in Khiops
        11.

    Parameters
    ----------
    dictionary_domain : `.DictionaryDomain`
        DictionaryDomain object. Its root dictionary must have its key set.
    root_dictionary_name : str
        Name for the new root dictionary
    secondary_table_variable_name : str
        Name, in the root dictionary, for the "table" variable of the secondary table.

    Returns
    -------
    `.DictionaryDomain`
        The new dictionary domain

    Raises
    ------
    `TypeError`
        Invalid type of an argument
    `ValueError`
        Invalid values of an argument:
        - the dictionary domain doesn't contain at least a dictionary
        - the dictionary domain's root dictionary doesn't have a key set
    """
    # Warn the user that this helper function is deprecated and will be removed
    warnings.warn(deprecation_message("build_multi_table_dictionary_domain", "11.0.0"))
    return _build_multi_table_dictionary_domain(
        dictionary_domain, root_dictionary_name, secondary_table_variable_name
    )


def _build_multi_table_dictionary_domain(
    dictionary_domain, root_dictionary_name, secondary_table_variable_name
):
    # Check that `dictionary_domain` is a `DictionaryDomain`
    if not isinstance(dictionary_domain, DictionaryDomain):
        raise TypeError(
            type_error_message("dictionary_domain", dictionary_domain, DictionaryDomain)
        )

    # Check input types
    if not isinstance(root_dictionary_name, str):
        raise TypeError(
            type_error_message("root_dictionary_name", root_dictionary_name, str)
        )
    if not isinstance(secondary_table_variable_name, str):
        raise TypeError(
            type_error_message(
                "secondary_table_variable_name", secondary_table_variable_name, str
            )
        )

    # Check that dictionary_domain has one dictionary (i.e. one table)
    if len(dictionary_domain.dictionaries) < 1:
        raise ValueError("'dictionary_domain' must contain at least one dictionary")

    # Get root source dictionary
    root_source_dictionary = dictionary_domain.dictionaries[0]

    # Check that root_source_dictionary has its key set:
    if not root_source_dictionary.key:
        raise ValueError("'root_source_dictionary' must have its key set")

    # Build and initialize root target dictionary from the input dictionary domain
    root_target_dictionary = Dictionary()
    root_target_dictionary.name = root_dictionary_name
    root_target_dictionary.key = root_source_dictionary.key
    root_target_dictionary.root = True

    # Copy the key variables from source dictionary to target dictionary
    for source_variable in root_source_dictionary.variables:
        if source_variable.name in root_source_dictionary.key:
            root_target_dictionary.add_variable(source_variable)

    # Build target variable for the target root dictionary
    target_variable = Variable()
    target_variable.name = secondary_table_variable_name
    target_variable.type = "Table"
    target_variable.object_type = root_source_dictionary.name
    root_target_dictionary.add_variable(target_variable)

    # Build secondary target dictionary, by copying root source dictionary
    secondary_target_dictionary = root_source_dictionary.copy()

    # Build target domain and add dictionaries to it
    target_domain = DictionaryDomain()
    target_domain.add_dictionary(root_target_dictionary)
    target_domain.add_dictionary(secondary_target_dictionary)

    return target_domain


# Disable the protected access rule because we need to call on the private API
# method for checking that a variable is a str or a DictionaryDomain
# pylint: disable=protected-access


def deploy_coclustering(
    dictionary_file_path_or_domain,
    dictionary_name,
    data_table_path,
    coclustering_file_path,
    key_variable_names,
    deployed_variable_name,
    results_dir,
    detect_format=True,
    header_line=None,
    field_separator=None,
    output_header_line=True,
    output_field_separator="\t",
    max_preserved_information=0,
    max_cells=0,
    max_part_numbers=None,
    build_cluster_variable=True,
    build_distance_variables=False,
    build_frequency_variables=False,
    variables_prefix="",
    results_prefix="",
    batch_mode=True,
    log_file_path=None,
    output_scenario_path=None,
    task_file_path=None,
    trace=False,
):
    r"""Deploys an *individual-variable* coclustering on a data table

    This procedure generates the following files in ``results_dir``:
        - ``Coclustering.kdic``: A multi-table dictionary file for further deployments
          of the coclustering with deploy_model
        - ``Keys<data_table_file_name>``: A data table file containing only the keys of
          individual
        - ``Deployed<data_table_file_name>``: A data table file containing the deployed
          coclustering model

    Parameters
    ----------
    dictionary_file_path_or_domain : str or `.DictionaryDomain`
        Path of a Khiops dictionary file or a DictionaryDomain object.
    dictionary_name : str
        Name of the dictionary to be analyzed.
    data_table_path : str
        Path of the data table file.
    coclustering_file_path : str
        Path of the coclustering model file (extension ``.khc`` or ``.khcj``)
    key_variable_names : list of str
        Names of the variables forming the unique keys of the individuals.
    deployed_variable_name : str
        Name of the coclustering variable to deploy.
    results_dir : str
        Path of the results directory.
    detect_format : bool, default ``True``
        If True detects automatically whether the data table file has a header and its
        field separator. It's ignored if ``header_line`` or ``field_separator`` are set.
    header_line : bool, optional (default ``True`` if ``detect_format`` is False)
        If True it uses the first line of the data as column names. Overrides
        ``detect_format`` if set.
    field_separator : str, optional (default "\\t" if ``detect_format`` is False)
        A field separator character, overrides ``detect_format`` if set ("" counts
        as "\\t").
    output_header_line : bool, default ``True``
        If True writes a header line containing the column names in the output table.
    output_field_separator : str, default "\\t"
        A field separator character (empty string counts as tab).
    max_preserved_information : int, default 0
        Maximum information preserve in the simplified coclustering. If equal to 0 there
        is no limit.
    max_cells : int, default 0
        Maximum number of cells in the simplified coclustering. If equal to 0 there is
        no limit.
    max_part_numbers : dict, optional
      Dictionary associating variable names to their maximum number of parts to
      preserve in the simplified coclustering. For variables not present in
      ``max_part_numbers`` there is no limit.
    build_cluster_variable : bool, default ``True``
        If True includes a cluster id variable in the deployment.
    build_distance_variables : bool, default False
        If True includes a cluster distance variable in the deployment.
    build_frequency_variables : bool, default False
        If True includes the frequency variables in the deployment.
    variables_prefix : str, default ""
        Prefix for the variables in the deployment dictionary.
    results_prefix : str, default ""
        Prefix of the result files.
    ... :
        Options of the `.KhiopsRunner.run` method from the class `.KhiopsRunner`.

    Returns
    -------
    tuple
        A 2-tuple containing:

        - The deployed data table path
        - The deployment dictionary file path.

    Raises
    ------
    `TypeError`
        Invalid type ``dictionary_file_path_or_domain`` or ``key_variable_names``
    `ValueError`
        If the type of the dictionary key variables is not equal to ``Categorical``

    Examples
    --------
    See the following function of the ``samples.py`` documentation script:
        - `samples.deploy_coclustering()`
    """
    # Obtain the dictionary of the table where the coclustering variables are
    api._check_dictionary_file_path_or_domain(dictionary_file_path_or_domain)
    if isinstance(dictionary_file_path_or_domain, DictionaryDomain):
        domain = dictionary_file_path_or_domain
    else:
        domain = read_dictionary_file(dictionary_file_path_or_domain)

    # Disambiguate the results directory path if necessary
    results_dir = create_unambiguous_khiops_path(results_dir)

    # Check the type of non basic keyword arguments specific to this function
    if not is_list_like(key_variable_names):
        raise TypeError(
            type_error_message("key_variable_names", key_variable_names, "ListLike")
        )

    # Detect the format once and for all to avoid inconsistencies
    if detect_format and header_line is None and field_separator is None:
        header_line, field_separator = api.detect_data_table_format(
            data_table_path, dictionary_file_path_or_domain, dictionary_name
        )
    else:
        if header_line is None:
            header_line = True
        if field_separator is None:
            field_separator = "\t"

    # Access the dictionary in the relevant variables
    dictionary = domain.get_dictionary(dictionary_name)

    # Verify that the key variables are categorical
    for key_variable_name in key_variable_names:
        key_variable = dictionary.get_variable(key_variable_name)
        if key_variable.type != "Categorical":
            raise ValueError(
                "key variable types must be 'Categorical', "
                f"variable '{key_variable_name}' has type '{key_variable.type}'"
            )
    # Make a copy of the dictionary and set the id_variable as key
    tmp_dictionary = dictionary.copy()
    tmp_dictionary.key = key_variable_names
    tmp_domain = DictionaryDomain()
    tmp_domain.add_dictionary(tmp_dictionary)

    # Create a root dictionary containing the keys
    root_dictionary_name = "CC_" + dictionary_name
    table_variable_name = "Table_" + dictionary_name
    domain = _build_multi_table_dictionary_domain(
        tmp_domain, root_dictionary_name, table_variable_name
    )

    # Create the deployment dictionary
    api.prepare_coclustering_deployment(
        domain,
        root_dictionary_name,
        coclustering_file_path,
        table_variable_name,
        deployed_variable_name,
        results_dir,
        max_preserved_information=max_preserved_information,
        max_cells=max_cells,
        max_part_numbers=max_part_numbers,
        build_cluster_variable=build_cluster_variable,
        build_distance_variables=build_distance_variables,
        build_frequency_variables=build_frequency_variables,
        variables_prefix=variables_prefix,
        results_prefix=results_prefix,
        batch_mode=batch_mode,
        log_file_path=log_file_path,
        output_scenario_path=output_scenario_path,
        task_file_path=task_file_path,
        trace=trace,
    )

    # Extract the keys from the tables to a temporary file
    data_table_file_name = os.path.basename(data_table_path)
    keys_table_file_path = fs.get_child_path(results_dir, f"Keys{data_table_file_name}")
    api.extract_keys_from_data_table(
        domain,
        dictionary_name,
        data_table_path,
        keys_table_file_path,
        header_line=header_line,
        field_separator=field_separator,
        output_header_line=header_line,
        output_field_separator=field_separator,
        trace=trace,
    )

    # Deploy the coclustering model
    coclustering_dictionary_file_path = fs.get_child_path(
        results_dir, "Coclustering.kdic"
    )
    output_data_table_path = fs.get_child_path(
        results_dir, f"Deployed{data_table_file_name}"
    )
    additional_data_tables = {
        f"{root_dictionary_name}`{table_variable_name}": data_table_path
    }
    api.deploy_model(
        coclustering_dictionary_file_path,
        root_dictionary_name,
        keys_table_file_path,
        output_data_table_path,
        header_line=header_line,
        field_separator=field_separator,
        output_header_line=output_header_line,
        output_field_separator=output_field_separator,
        additional_data_tables=additional_data_tables,
        trace=trace,
    )
    return output_data_table_path, coclustering_dictionary_file_path


def deploy_predictor_for_metrics(
    dictionary_file_path_or_domain,
    dictionary_name,
    data_table_path,
    output_data_table_path,
    detect_format=True,
    header_line=None,
    field_separator=None,
    sample_percentage=70,
    sampling_mode="Include sample",
    additional_data_tables=None,
    output_header_line=True,
    output_field_separator="\t",
    trace=False,
):
    r"""Deploys the necessary data to estimate the performance metrics of a predictor

    For each instance for each instance it deploys:

    - The true value of the target variable
    - The predicted value of the target variable
    - The probabilities of each value of the target variable *(classifier only)*

    .. note::
        To obtain the data of the default Khiops test dataset use ``sample_percentage =
        70`` and ``sampling_mode = "Exclude sample"``.

    Parameters
    ----------
    dictionary_file_path_or_domain : str or `.DictionaryDomain`
        Path of a Khiops dictionary file or a DictionaryDomain object.
    dictionary_name : str
        Name of the predictor dictionary.
    data_table_path : str
        Path of the data table file.
    output_data_table_path : str
        Path of the scores output data file.
    detect_format : bool, default ``True``
        If True detects automatically whether the data table file has a header and its
        field separator. It's ignored if ``header_line`` or ``field_separator`` are set.
    header_line : bool, optional (default ``True`` if ``detect_format`` is ``False``)
        If True it uses the first line of the data as column names. Overrides
        ``detect_format`` if set.
    field_separator : str, optional (default "\\t" if ``detect_format`` is ``False``)
        A field separator character, overrides ``detect_format`` if set ("" counts
        as "\\t").
    sample_percentage : int, default 70
        See ``sampling_mode`` option below.
    sampling_mode : "Include sample" or "Exclude sample"
        If equal to "Include sample" deploys the predictor on ``sample_percentage``
        percent of data and if equal to "Exclude sample" on the complementary ``100 -
        sample_percentage`` percent of data.
    additional_data_tables : dict, optional
        A dictionary containing the data paths and file paths for a multi-table
        dictionary file. For more details see :doc:`/multi_table_primer` documentation.
    output_header_line : bool, default ``True``
        If True writes a header line containing the column names in the output table.
    output_field_separator : str, default "\\t"
        A field separator character ("" counts as "\\t").
    ... :
        Options of the `.KhiopsRunner.run` method from the class `.KhiopsRunner`.
    """
    # Check the dictionary domain
    api._check_dictionary_file_path_or_domain(dictionary_file_path_or_domain)

    # Load the dictionary file into a domain if necessary
    if isinstance(dictionary_file_path_or_domain, DictionaryDomain):
        predictor_domain = dictionary_file_path_or_domain.copy()
    else:
        predictor_domain = read_dictionary_file(dictionary_file_path_or_domain)

    # Check that the specified dictionary is a predictor
    predictor_dictionary = predictor_domain.get_dictionary(dictionary_name)
    if "PredictorType" not in predictor_dictionary.meta_data:
        raise ValueError(f"Dictionary '{predictor_dictionary.name}' is not a predictor")

    # Set the type of classifier
    predictor_type = predictor_dictionary.meta_data.get_value("PredictorType")
    is_classifier = predictor_type == "Classifier"

    # Use the necessary columns
    predictor_dictionary.use_all_variables(False)
    for variable in predictor_dictionary.variables:
        if "TargetVariable" in variable.meta_data:
            variable.used = True
        elif is_classifier:
            if "Prediction" in variable.meta_data:
                variable.used = True
            for key in variable.meta_data.keys:
                if key.startswith("TargetProb"):
                    variable.used = True
        elif not is_classifier and "Mean" in variable.meta_data:
            variable.used = True

    # Deploy the scores
    api.deploy_model(
        predictor_domain,
        dictionary_name,
        data_table_path,
        output_data_table_path,
        detect_format=detect_format,
        header_line=header_line,
        field_separator=field_separator,
        sample_percentage=sample_percentage,
        sampling_mode=sampling_mode,
        additional_data_tables=additional_data_tables,
        output_header_line=output_header_line,
        output_field_separator=output_field_separator,
        trace=trace,
    )


# pylint: enable=protected-access


def visualize_report(report_path):
    """Opens a Khiops or Khiops Coclustering report with the desktop visualization app

    Before using this function, make sure you have installed the Khiops Visualization
    app and/or the Khiops Co-Visualization app. More info at
    `<https://khiops.org/setup/visualization/>`_

    Parameters
    ----------
    report_path : str
        The path of the report file to be open. It must have extension '.khj' (Khiops
        report) or '.khcj' (Khiops Coclustering report).

    Raises
    ------
    `ValueError`
        If the report file path does not have extension '.khj' or '.khcj'.
    `FileNotFoundError`
        If the report file does not exist.
    `RuntimeError`
        If the report file is executable.
    """
    # Check that the report path:
    # - has a valid file extension
    # - exists
    # - is not executable
    #   - Skip this check on Windows because generated reports are executable
    _, ext = os.path.splitext(report_path)
    if ext not in [".khj", ".khcj"]:
        raise ValueError(
            "'report_path' must have extension '.khj' or '.khcj'. "
            f"Path: {report_path}"
        )
    if not os.path.exists(report_path):
        raise FileNotFoundError(report_path)
    if platform.system() != "Windows" and os.access(report_path, os.X_OK):
        raise RuntimeError(f"Report file cannot be executable. Path: {report_path}")

    # Open it with the associated application
    try:
        if platform.system() == "Windows":
            subprocess.call(["explorer", report_path])
        elif platform.system() == "Darwin":
            subprocess.call(["open", report_path])
        else:
            subprocess.call(["xdg-open", report_path])
    # On failure we just print the error to not break the execution
    except OSError as error:
        print(f"Could not open report file: {error}. Path: {report_path}")



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""train_predictor task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import (
    BoolType,
    DictType,
    FloatType,
    IntType,
    ListType,
    StringLikeType,
    TupleType,
)

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "train_predictor",
        "khiops",
        "10.0.0",
        [
            ("dictionary_file_path", StringLikeType),
            ("dictionary_name", StringLikeType),
            ("data_table_path", StringLikeType),
            ("target_variable", StringLikeType),
            ("results_dir", StringLikeType),
        ],
        [
            ("detect_format", BoolType, True),
            ("header_line", BoolType, True),
            ("field_separator", StringLikeType, ""),
            ("sample_percentage", FloatType, 100.0),
            ("sampling_mode", StringLikeType, "Include sample"),
            ("test_database_mode", StringLikeType, "Complementary"),
            ("selection_variable", StringLikeType, ""),
            ("selection_value", StringLikeType, ""),
            ("additional_data_tables", DictType(StringLikeType, StringLikeType), None),
            ("main_target_value", StringLikeType, ""),
            ("snb_predictor", BoolType, True),
            ("univariate_predictor_number", IntType, 0),
            ("max_evaluated_variables", IntType, 0),
            ("max_selected_variables", IntType, 0),
            ("max_constructed_variables", IntType, 0),
            ("construction_rules", ListType(StringLikeType), None),
            ("max_trees", IntType, 10),
            ("max_pairs", IntType, 0),
            ("all_possible_pairs", BoolType, True),
            (
                "specific_pairs",
                ListType(TupleType(StringLikeType, StringLikeType)),
                None,
            ),
            ("group_target_value", BoolType, False),
            ("discretization_method", StringLikeType, "MODL"),
            ("min_interval_frequency", IntType, 0),
            ("max_intervals", IntType, 0),
            ("grouping_method", StringLikeType, "MODL"),
            ("min_group_frequency", IntType, 0),
            ("max_groups", IntType, 0),
            ("results_prefix", StringLikeType, ""),
        ],
        [
            "dictionary_file_path",
            "data_table_path",
            "results_dir",
            "additional_data_tables",
        ],
        # pylint: disable=line-too-long
        # fmt: off
        """
        // Dictionary file and class settings
        ClassManagement.OpenFile
        ClassFileName __dictionary_file_path__
        OK
        ClassManagement.ClassName __dictionary_name__

        // Train/test database settings
        TrainDatabase.DatabaseFiles.List.Key __dictionary_name__
        TrainDatabase.DatabaseFiles.DataTableName __data_table_path__
        __DICT__
        __additional_data_tables__
        TrainDatabase.DatabaseFiles.List.Key
        TrainDatabase.DatabaseFiles.DataTableName
        __END_DICT__
        TrainDatabase.HeaderLineUsed __header_line__
        TrainDatabase.FieldSeparator __field_separator__
        __OPT__
        __detect_format__
        TrainDatabase.DatabaseFormatDetector.DetectFileFormat
        __END_OPT__
        TrainDatabase.SampleNumberPercentage __sample_percentage__
        TrainDatabase.SamplingMode __sampling_mode__
        TrainDatabase.SelectionAttribute __selection_variable__
        TrainDatabase.SelectionValue __selection_value__
        TrainDatabase.TestDatabaseSpecificationMode __test_database_mode__

        // Target variable
        AnalysisSpec.TargetAttributeName __target_variable__
        AnalysisSpec.MainTargetModality __main_target_value__

        // Predictors to train
        AnalysisSpec.PredictorsSpec.SelectiveNaiveBayesPredictor __snb_predictor__
        AnalysisSpec.PredictorsSpec.AdvancedSpec.UnivariatePredictorNumber __univariate_predictor_number__

        // Selective Naive Bayes settings
        AnalysisSpec.PredictorsSpec.AdvancedSpec.InspectSelectiveNaiveBayesParameters
        TrainParameters.MaxEvaluatedAttributeNumber __max_evaluated_variables__
        SelectionParameters.MaxSelectedAttributeNumber __max_selected_variables__
        Exit

        // Feature engineering
        AnalysisSpec.PredictorsSpec.ConstructionSpec.MaxTreeNumber __max_trees__
        AnalysisSpec.PredictorsSpec.ConstructionSpec.MaxAttributePairNumber __max_pairs__
        AnalysisSpec.PredictorsSpec.AdvancedSpec.InspectAttributePairsParameters
        AllAttributePairs __all_possible_pairs__
        __LIST__
        __specific_pairs__
        SpecificAttributePairs.InsertItemAfter
        SpecificAttributePairs.FirstName
        SpecificAttributePairs.SecondName
        __END_LIST__
        Exit
        AnalysisSpec.PredictorsSpec.ConstructionSpec.MaxConstructedAttributeNumber __max_constructed_variables__
        AnalysisSpec.PredictorsSpec.AdvancedSpec.InspectConstructionDomain
        __DICT__
        __construction_rules__
        UnselectAll
        ConstructionRules.List.Key
        ConstructionRules.Used
        __END_DICT__
        Exit

        // Data preparation (discretization & grouping) settings
        AnalysisSpec.PreprocessingSpec.TargetGrouped __group_target_value__
        AnalysisSpec.PreprocessingSpec.DiscretizerSpec.SupervisedMethodName __discretization_method__
        AnalysisSpec.PreprocessingSpec.DiscretizerSpec.UnsupervisedMethodName __discretization_method__
        AnalysisSpec.PreprocessingSpec.DiscretizerSpec.MinIntervalFrequency __min_interval_frequency__
        AnalysisSpec.PreprocessingSpec.DiscretizerSpec.MaxIntervalNumber __max_intervals__
        AnalysisSpec.PreprocessingSpec.GrouperSpec.SupervisedMethodName __grouping_method__
        AnalysisSpec.PreprocessingSpec.GrouperSpec.UnsupervisedMethodName __grouping_method__
        AnalysisSpec.PreprocessingSpec.GrouperSpec.MinGroupFrequency __min_group_frequency__
        AnalysisSpec.PreprocessingSpec.GrouperSpec.MaxGroupNumber __max_groups__

        // Output settings
        AnalysisResults.ResultFilesDirectory __results_dir__
        AnalysisResults.ResultFilesPrefix __results_prefix__

        // Build model
        ComputeStats
        """,
        # fmt: on
    ),
]



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Classes for creating Khiops scenario files"""
import re

from khiops.core.internals.common import is_string_like


class ConfigurableKhiopsScenario:
    """A configurable Khiops scenario

    This class encapsulates a template Khiops scenario and its parameters. It allows to
    replace the template keyword to write an executable scenario.

    Parameters
    ----------
    template : str
        The template scenario.

    Attributes
    ----------
    template : str
        The template scenario file.
    """

    def __init__(self, template):
        """See class docstring"""
        # Save and parse the template
        self.template = template
        self._parse_template()

    def _parse_template(self):
        # Initialize private fields
        self._parsed_template = []

        # Matches a template entry
        template_entry_re = re.compile(r"\s*(([a-zA-Z.]+)\s+)?(__\w+__)")

        # Iterate over the template lines
        line_iter = iter(self.template.split("\n"))
        line = next(line_iter, None)
        line_num = 1
        while line is not None:
            entry = None
            match = template_entry_re.fullmatch(line)
            # If there is no match add the line as-is
            if match is None:
                entry = line
            # Otherwise apply the detect process depending of the entry type
            else:
                field_name = match.group(2)
                keyword = match.group(3)
                if field_name is not None:
                    param_name = keyword
                    entry = (param_name, field_name)
                else:
                    if keyword == "__OPT__":
                        entry = self._parse_opt(line_iter)
                        _, _, statements = entry
                        line_num += 2 + len(statements)
                    elif keyword == "__LIST__":
                        entry = self._parse_list(line_iter)
                        _, _, prologue, _, entry_values = entry
                        line_num += 3 + len(prologue) + len(entry_values)
                    elif keyword == "__DICT__":
                        entry = self._parse_dict(line_iter)
                        _, _, prologue, *_ = entry
                        line_num += 4 + len(prologue)
                    else:
                        raise ValueError(
                            "Expected keyword __DICT__, __OPT__ "
                            f"or __LIST__ not '{keyword}'"
                        )
            assert entry is not None, "entry must not be None"
            self._parsed_template.append(entry)

            # Update the template line
            line = next(line_iter, None)
            line_num += 1

        # Eliminate empty lines at the beginning and end if any
        if self._parsed_template[0] == "":
            self._parsed_template.pop(0)
        if self._parsed_template[-1] == "":
            self._parsed_template.pop()

        assert isinstance(self._parsed_template, list)

    def _parse_section(self, section_keyword, line_iter):
        # Obtain the end keyword
        end_section_keyword = f"__END_{section_keyword[2:]}"
        end_section_re = re.compile(r"\s*" + end_section_keyword)

        # Obtain the section lines
        section_spec = []
        match = None
        while match is None:
            try:
                line = next(line_iter)
                match = end_section_re.fullmatch(line)
            except StopIteration as error:
                raise ValueError(
                    f"{section_keyword} section has "
                    f"no matching {end_section_keyword}"
                ) from error
            section_spec.append(line)

        # Eliminate the last line which is necessarily the end keyword
        section_spec.pop()

        # Separate the parameter name from the spec
        section_param_name = section_spec.pop(0).lstrip()

        # Check that the parameter name conforms to the parameter naming
        param_name_re = re.compile(r"__\w+__")
        if param_name_re.match(section_param_name) is None:
            raise ValueError(
                f"{section_keyword} template parameter name does not conform "
                f"to the __param__ notation: '{section_param_name}'"
            )
        # Check that the section spec conforms to the scenario commands naming
        statement_re = re.compile(r"\s*[a-zA-Z][a-zA-Z.]*(\s+.*)?$")
        for statement in section_spec:
            if statement_re.match(statement) is None:
                raise ValueError(
                    "Statement must contain only alphabetic characters and '.' "
                    f"(no '.' at the beginning): '{statement}'"
                )

        return section_param_name, section_spec

    def _parse_opt(self, line_iter):
        opt_param_name, opt_section_spec = self._parse_section("__OPT__", line_iter)
        return (opt_param_name, "__OPT__", opt_section_spec)

    def _parse_list(self, line_iter):
        list_param_name, list_spec = self._parse_section("__LIST__", line_iter)

        # Check the section length
        if len(list_spec) < 2:
            raise ValueError("__LIST__ section must have at least 3 statements")

        # Look for the InsertItemAfter statement index
        list_value_start = None
        for i, line in enumerate(list_spec):
            if line.endswith("InsertItemAfter"):
                list_value_start = line
                list_value_start_index = i
                break
        if list_value_start is None:
            raise ValueError(
                "__LIST__ section does not contain "
                "list statement ending with '.InsertItemAfter'"
            )
        if list_value_start_index == len(list_spec) - 1:
            raise ValueError("__LIST__ section does not contain any value statement")

        # Separate into:
        # - the prologue
        # - the tuple start statement
        # - the tuple value statements
        list_prologue = list_spec[:list_value_start_index]
        list_value_start = list_spec[list_value_start_index]
        list_value_names = list_spec[list_value_start_index + 1 :]

        return (
            list_param_name,
            "__LIST__",
            list_prologue,
            list_value_start,
            list_value_names,
        )

    def _parse_dict(self, line_iter):
        # Obtain the parsed elements from the line iterator
        dict_param_name, dict_section = self._parse_section("__DICT__", line_iter)

        # Separate into
        # - The prologue statements
        # - The key statement
        # - The value statement
        if len(dict_section) < 2:
            raise ValueError("__DICT__ section must have at least 2 statements")
        dict_prologue = dict_section[:-2]
        dict_key_field_name = dict_section[-2]
        dict_value_field_name = dict_section[-1]
        if not dict_key_field_name.endswith(".Key"):
            raise ValueError("__DICT__ key statement must end with '.Key'")

        return (
            dict_param_name,
            "__DICT__",
            dict_prologue,
            dict_key_field_name,
            dict_value_field_name,
        )

    def __str__(self):
        return str(self._parsed_template)

    def check_keyword_completeness(self, call_keywords):
        """Check if a list of keys is exactly that of the template of this scenario"""
        # Transform keywords to set
        call_keyword_set = set(call_keywords)
        template_keyword_set = set(
            entry[0] for entry in self._parsed_template if isinstance(entry, tuple)
        )

        # Check if there are unknown keys in the input keywords
        extra_call_keyword_set = call_keyword_set - template_keyword_set
        if extra_call_keyword_set:
            raise ValueError(
                "Call parameter keyword(s) not in template keywords '"
                + ",'".join(extra_call_keyword_set)
                + "'"
            )

        # Check if there are missing keys in the input keywords
        extra_template_keyword_set = template_keyword_set - call_keyword_set
        if extra_template_keyword_set:
            raise ValueError(
                "Template parameter keyword(s) not in call keywords '"
                + ",'".join(extra_template_keyword_set)
                + "'"
            )

    def write(self, writer, scenario_args):
        """Writes the de-templatized scenario with the specified arguments

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            The output writer for the scenario file.
        scenario_args : dict
            Values of the scenario template arguments.
        """
        for entry in self._parsed_template:
            if isinstance(entry, str):
                writer.writeln(entry)
            else:
                assert isinstance(entry, tuple)
                assert len(entry) >= 2
                entry_arg_name, entry_type, *_ = entry
                arg_value = scenario_args[entry_arg_name]
                if entry_type not in ["__DICT__", "__OPT__", "__LIST__"]:
                    self._write_str(writer, entry, arg_value)
                elif entry_type == "__OPT__":
                    self._write_opt(writer, entry, arg_value)
                elif entry_type == "__DICT__" and arg_value is not None:
                    self._write_dict(writer, entry, arg_value)
                elif entry_type == "__LIST__" and arg_value is not None:
                    self._write_list(writer, entry, arg_value)

    def _write_str(self, writer, entry, arg_value):
        assert is_string_like(arg_value)
        assert len(entry) == 2
        _, entry_name = entry
        writer.write(f"{entry_name} ")
        writer.writeln(arg_value)

    def _write_opt(self, writer, entry, arg_value):
        assert isinstance(arg_value, str)
        assert len(entry) == 3
        _, entry_type, opt_actions = entry
        assert entry_type == "__OPT__"
        if arg_value == "true":
            for action in opt_actions:
                writer.writeln(action)

    def _write_dict(self, writer, entry, arg_values):
        assert isinstance(arg_values, (dict, list))
        assert len(entry) == 5

        # Unroll the entry
        _, entry_type, entry_prologue, entry_key, entry_value = entry
        assert entry_type == "__DICT__"

        # Write the prologue
        for line in entry_prologue:
            writer.writeln(line)

        # Transform list arguments to "true"-valued dictionary
        if isinstance(arg_values, list):
            dict_arg_values = {key: "true" for key in arg_values}
        else:
            dict_arg_values = arg_values

        # Write the argument values
        for key, value in dict_arg_values.items():
            writer.write(f"{entry_key} ")
            writer.writeln(key)
            writer.write(f"{entry_value} ")
            writer.writeln(value)

    def _write_list(self, writer, entry, arg_values):
        assert isinstance(arg_values, list)
        for value in arg_values:
            assert is_string_like(value) or isinstance(value, tuple)

        # Unroll the entry
        _, entry_type, entry_prologue, entry_header, entry_value_names = entry
        assert entry_type == "__LIST__"

        # Write the prologue
        for line in entry_prologue:
            writer.writeln(line)

        # Write the contained values of the list argument
        for value in arg_values:
            writer.writeln(entry_header)

            # Tuple list case
            if isinstance(value, tuple):
                assert len(entry_value_names) == len(value)
                for entry_value_name, subvalue in zip(entry_value_names, value):
                    writer.write(f"{entry_value_name} ")
                    writer.writeln(subvalue)
            # String list case
            else:
                assert len(entry_value_names) == 1
                assert is_string_like(value)
                writer.write(f"{entry_value_names[0]} ")
                writer.writeln(value)



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Classes to handle Khiops specific I/O"""
import copy
import io
import json
import os
import platform
import warnings

import khiops.core.internals.filesystems as fs
from khiops.core.exceptions import KhiopsJSONError
from khiops.core.internals.common import is_string_like, type_error_message


def encode_file_path(file_path):
    """Encodes a file path

    This is custom path encoding for Khiops scenarios that is platform dependent. The
    encoding is done only if file_path is of type str.

    Parameters
    ----------
    file_path : str or bytes
        The path of a file.

    Returns
    -------
    `bytes`
        If ``file_path`` is str
            - In Windows : The path decoded to UTF-8 excepting the "ANSI" Unicode
              characters.
            - In Linux/Unix/Mac : The path decoded to UTF-8.
        If ``file_path`` is `bytes`:
            It just returns the input ``file_path``

    Raises
    ------
    `TypeError`
        If ``file_path`` is not str or bytes
    """
    # Check input type
    if not is_string_like(file_path):
        raise TypeError(type_error_message("file_path", file_path, str, bytes))

    # Return as-is if it is a byte sequence
    if not isinstance(file_path, str):
        assert isinstance(file_path, bytes)
        return file_path

    # Platform *nix: return UTF-8 encoded path
    if platform.system() != "Windows":
        return bytes(file_path, encoding="utf8")

    # Platform Windows:
    # - Return ANSI encoded chars if they over the 128-255 window
    # - Return UTF8 encoded chars otherwise
    decoded_bytes = bytearray()
    for char in file_path:
        if char in all_ansi_unicode_chars:
            decoded_bytes.extend(all_ansi_unicode_chars_to_ansi[char])
        else:
            decoded_bytes.extend(bytearray(char, "utf8"))
    return bytes(decoded_bytes)


def flexible_json_load(json_file_path):
    """Loads flexibly a JSON file

    First it tries a vanilla read, then if that fails it warns and then loads the files
    replacing the errors.

    Parameters
    ----------
    json_file_path : str
        Path of the Khiops JSON file.

    Returns
    -------
    dict
        The in-memory representation of the JSON file.
    """
    with io.BytesIO(fs.read(json_file_path)) as json_file_stream:
        try:
            json_data = json.load(json_file_stream)
            first_load_failed = False
        except UnicodeDecodeError as error:
            warnings.warn(
                "Khiops JSON file raised UnicodeDecodeError, "
                "probably because the file is not encoded in UTF-8. "
                "The file will be loaded with replacement characters on "
                "decoding errors and this may generate problems downstream. "
                "To avoid any problem, regenerate the file with Khiops 10.0 "
                f"or newer.\nKhiops JSON file: {json_file_path}.\n"
                f"UnicodeDecodeError message:\n{error}"
            )
            first_load_failed = True

    # Try a second time with flexible errors
    if first_load_failed:
        with io.StringIO(
            fs.read(json_file_path).decode("utf8", errors="replace")
        ) as json_file_stream:
            json_data = json.load(json_file_stream)

    return json_data


class KhiopsJSONObject:
    """Represents the contents of a Khiops JSON file

    Parameters
    ----------
    json_data : dict, optional
        Python dictionary representing the data of a Khiops JSON file. If None an empty
        it returns an empty object.

    Raises
    ------
    `~.KhiopsJSONError`
        If the JSON data is invalid.

    Attributes
    ----------
    tool : str
        Name of the Khiops tool that generated the file.
    version : str
        Version of the Khiops tool that generated the file.
    khiops_encoding : str, optional
        Custom encoding used by Khiops in the file. Valid values:
            - ``None`` : for backwards compatibility
            - "ascii": ASCII encoding
            - "ansi": ANSI encoding
            - "utf8": UTF-8 encoding
            - "mixed_ansi_utf8": Mixed characters from UTF-8 and ANSI but no collision.
            - "colliding_ansi_utf8" : Colliding characters from UTF-8 and ANSI.
    sub_tool : str, optional
        Identifies the tool that originated the JSON file. Used by tools of the Khiops
        family such as PataText or Enneade.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Check the type of json_data
        if json_data is not None and not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Input check
        if json_data is None:
            json_data = {}
        else:
            if "tool" not in json_data:
                raise KhiopsJSONError("Khiops JSON file does not have a 'tool' field")
            if "version" not in json_data:
                raise KhiopsJSONError(
                    "Khiops JSON file does not have a 'version' field"
                )
            if "khiops_encoding" not in json_data:
                warnings.warn(
                    "Khiops JSON file does not have a 'khiops_encoding' field "
                    "(generated with Khiops older than 10.0.1?). "
                    "Exported dictionary/report files may be corrupted. ",
                    stacklevel=6,
                )
            elif json_data["khiops_encoding"] == "colliding_ansi_utf8":
                if "colliding_utf8_chars" in json_data:
                    colliding_chars_message = "Colliding chars: " + ", ".join(
                        json_data["colliding_utf8_chars"]
                    )
                else:
                    colliding_chars_message = ""
                warnings.warn(
                    "Khiops JSON file contains colliding characters. "
                    "Exported dictionary/report files may be unusable. "
                    f"{colliding_chars_message}",
                    stacklevel=6,
                )
            elif json_data["khiops_encoding"] not in [
                "ascii",
                "ansi",
                "utf8",
                "mixed_ansi_utf8",
            ]:
                raise KhiopsJSONError(
                    "Khiops JSON file 'khiops_encoding' field value must be 'ascii', "
                    "'ansi', 'utf8', 'mixed_ansi_utf8' or 'colliding_ansi_utf8', "
                    f"""not '{json_data["khiops_encoding"]}'."""
                )

        # Initialize attributes from data
        # `tool` and `version` need to be strings by default, so that they can
        # be written
        self.tool = json_data.get("tool", "")
        self.version = json_data.get("version", "")
        self.sub_tool = json_data.get("subTool")

        # Obtain encoding fields
        self.khiops_encoding = json_data.get("khiops_encoding")
        if self.khiops_encoding is not None:
            self.ansi_chars = json_data.get("ansi_chars")
            self.colliding_utf8_chars = json_data.get("colliding_utf8_chars")
        # To support Khiops < 10.0.1
        else:
            self.ansi_chars = []
            self.colliding_utf8_chars = []

        # Store a copy of the data to be able to write copies of it
        self.json_data = copy.deepcopy(json_data)

    def create_output_file_writer(self, stream):
        """Creates an output file with the proper encoding settings

        Parameters
        ----------
        stream : `io.IOBase`
            An output stream object.

        Returns
        -------
        `.KhiopsOutputWriter`
            An output file object.
        """
        if self.khiops_encoding is None or self.khiops_encoding in ["ascii", "utf8"]:
            return KhiopsOutputWriter(stream)
        else:
            if self.khiops_encoding == "colliding_ansi_utf8":
                return KhiopsOutputWriter(
                    stream, force_ansi=True, ansi_unicode_chars=self.ansi_chars
                )
            else:
                return KhiopsOutputWriter(stream, force_ansi=True)

    def write_khiops_json_file(self, json_file_path):
        """Write the JSON data of the object to a Khiops JSON file

        Parameters
        ----------
        json_file_path : str
            Path to the Khiops JSON file.
        """
        if self.json_data is not None:
            # Serialize JSON data to string
            # Do not escape non-ASCII Unicode characters
            json_string = json.dumps(self.json_data, ensure_ascii=False)
            with io.BytesIO() as json_stream:
                writer = self.create_output_file_writer(json_stream)
                writer.write(json_string)
                fs.write(uri_or_path=json_file_path, data=json_stream.getvalue())


class KhiopsOutputWriter:
    """Output writer with additional services to handle Khiops special encodings

    Parameters
    ----------
    stream : `io.IOBase`
        A writable output stream. Special text transformations in buffers inheriting
        from `io.TextIOBase` are ignored.
    force_ansi : bool, default False
        All output written will be transformed back ANSI characters in that range that
        were recoded to UTF-8.
    ansi_unicode_chars : list of str, optional
        A list of UTF-8 characters with equivalents in the ANSI 128-256 range which will
        be encoded back to ANSI when writing with ``force_ansi`` is ``True``. By default
        all UTF-8 equivalents of the ANSI 128-256 will be encoded back.
    """

    def __init__(self, stream, force_ansi=False, ansi_unicode_chars=None):
        # Set the output stream
        if isinstance(stream, io.IOBase):
            if not stream.writable():
                raise ValueError("'stream' must be writable")
            if isinstance(stream, io.TextIOBase):
                self.stream = stream.buffer
            else:
                self.stream = stream
        else:
            raise ValueError(type_error_message("stream", stream, io.IOBase))

        # Set the force_ansi parameter
        self.force_ansi = force_ansi
        if ansi_unicode_chars is None:
            self.ansi_unicode_chars = all_ansi_unicode_chars
        else:
            for char in ansi_unicode_chars:
                if char not in all_ansi_unicode_chars:
                    raise ValueError(
                        f"Unicode char '{char}' does not have an equivalent "
                        "in the 128-256 ANSI range"
                    )
            self.ansi_unicode_chars = ansi_unicode_chars

    def write(self, string):
        if not is_string_like(string):
            raise TypeError(type_error_message("string", string, "string-like"))

        if isinstance(string, str):
            if self.force_ansi:
                self._write_ansi(string)
            else:
                self.stream.write(bytes(string, "utf8"))
        else:
            self.stream.write(string)

    def writeln(self, string):
        self.write(string)
        self.write(os.linesep)

    def close(self):
        self.stream.close()

    def _write_ansi(self, string):
        for char in string:
            if char in self.ansi_unicode_chars:
                self.stream.write(all_ansi_unicode_chars_to_ansi[char])
            else:
                self.stream.write(bytes(char, "utf8"))


# Mapping ANSI -> Unicode as a list of 128 Unicode characters.
# Only non-ASCII characters are represented, that is the range 128-255.
all_ansi_unicode_chars = [
    "\u20AC",
    "\u0081",
    "\u201A",
    "\u0192",
    "\u201E",
    "\u2026",
    "\u2020",
    "\u2021",
    "\u02C6",
    "\u2030",
    "\u0160",
    "\u2039",
    "\u0152",
    "\u008D",
    "\u017D",
    "\u008F",
    "\u0090",
    "\u2018",
    "\u2019",
    "\u201C",
    "\u201D",
    "\u2022",
    "\u2013",
    "\u2014",
    "\u02DC",
    "\u2122",
    "\u0161",
    "\u203A",
    "\u0153",
    "\u009D",
    "\u017E",
    "\u0178",
    "\u00A0",
    "\u00A1",
    "\u00A2",
    "\u00A3",
    "\u00A4",
    "\u00A5",
    "\u00A6",
    "\u00A7",
    "\u00A8",
    "\u00A9",
    "\u00AA",
    "\u00AB",
    "\u00AC",
    "\u00AD",
    "\u00AE",
    "\u00AF",
    "\u00B0",
    "\u00B1",
    "\u00B2",
    "\u00B3",
    "\u00B4",
    "\u00B5",
    "\u00B6",
    "\u00B7",
    "\u00B8",
    "\u00B9",
    "\u00BA",
    "\u00BB",
    "\u00BC",
    "\u00BD",
    "\u00BE",
    "\u00BF",
    "\u00C0",
    "\u00C1",
    "\u00C2",
    "\u00C3",
    "\u00C4",
    "\u00C5",
    "\u00C6",
    "\u00C7",
    "\u00C8",
    "\u00C9",
    "\u00CA",
    "\u00CB",
    "\u00CC",
    "\u00CD",
    "\u00CE",
    "\u00CF",
    "\u00D0",
    "\u00D1",
    "\u00D2",
    "\u00D3",
    "\u00D4",
    "\u00D5",
    "\u00D6",
    "\u00D7",
    "\u00D8",
    "\u00D9",
    "\u00DA",
    "\u00DB",
    "\u00DC",
    "\u00DD",
    "\u00DE",
    "\u00DF",
    "\u00E0",
    "\u00E1",
    "\u00E2",
    "\u00E3",
    "\u00E4",
    "\u00E5",
    "\u00E6",
    "\u00E7",
    "\u00E8",
    "\u00E9",
    "\u00EA",
    "\u00EB",
    "\u00EC",
    "\u00ED",
    "\u00EE",
    "\u00EF",
    "\u00F0",
    "\u00F1",
    "\u00F2",
    "\u00F3",
    "\u00F4",
    "\u00F5",
    "\u00F6",
    "\u00F7",
    "\u00F8",
    "\u00F9",
    "\u00FA",
    "\u00FB",
    "\u00FC",
    "\u00FD",
    "\u00FE",
    "\u00FF",
]

# Map the ANSI 128-255 range coded in Unicode back to its ANSI encoding
all_ansi_unicode_chars_to_ansi = {
    char: bytes([128 + i]) for i, char in enumerate(all_ansi_unicode_chars)
}



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""detect_data_table_format_with_dictionary task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import StringLikeType

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "detect_data_table_format_with_dictionary",
        "khiops",
        "10.0.1",
        [
            ("dictionary_file_path", StringLikeType),
            ("dictionary_name", StringLikeType),
            ("data_table_path", StringLikeType),
        ],
        [],
        ["data_table_path"],
        # fmt: off
        """
        // Dictionary file and class settings
        ClassManagement.OpenFile
        ClassFileName __dictionary_file_path__
        OK
        ClassManagement.ClassName __dictionary_name__

        // Detect format the data table format on the "Extract Keys" window
        LearningTools.ExtractKeysFromDataTable
        SourceDataTable.DatabaseName __data_table_path__
        SourceDataTable.DatabaseFormatDetector.DetectFileFormat
        Exit
        """,
        # fmt: on
    ),
]



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Python package builder and installer driver
"""

from setuptools import find_packages, setup

import versioneer

if __name__ == "__main__":
    setup(
        name="khiops",
        version=versioneer.get_version(),
        url="https://khiops.org",
        description="Python library for the Khiops AutoML suite",
        maintainer="The Khiops Team",
        maintainer_email="khiops.team@orange.com",
        license_files=["LICENSE.md"],
        entry_points={
            "console_scripts": [
                "kh-status=khiops.tools:kh_status_entry_point",
                "kh-samples=khiops.tools:kh_samples_entry_point",
                "kh-download-datasets=khiops.tools:kh_download_datasets_entry_point",
                "pk-status=khiops.tools:pk_status_entry_point",  # deprecated
            ]
        },
        packages=find_packages(include=["khiops", "khiops.*"]),
        include_package_data=True,
        python_requires=">=3.8",
        install_requires=[
            "pandas>=0.25.3",
            "scikit-learn>=0.22.2",
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: Other/Proprietary License",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering",
        ],
        cmdclass=versioneer.get_cmdclass(),
        extras_require={
            "s3": [
                # do not necessary use the latest version
                # to avoid undesired breaking changes
                "boto3>=1.17.39,<=1.35.69",
                # an open issue on boto3 (https://github.com/boto/boto3/issues/3585)
                # forces a minimal version of pyopenssl
                "pyopenssl>=24.0.0,<25.0.0",
            ],
            "gcs": ["google-cloud-storage>=1.37.0"],
        },
    )



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Khiops Python exception classes"""


class KhiopsJSONError(Exception):
    """Parsing error for Khiops-generated JSON files"""


class KhiopsRuntimeError(Exception):
    """Khiops execution related errors"""


class KhiopsEnvironmentError(Exception):
    """Khiops execution environment error

    Example: Khiops binary not found.
    """



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""simplify_coclustering task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import DictType, IntType, StringLikeType

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "simplify_coclustering",
        "khiops_coclustering",
        "10.1.0",
        [
            ("coclustering_file_path", StringLikeType),
            ("simplified_coclustering_file_path", StringLikeType),
            ("results_dir", StringLikeType),
        ],
        [
            ("max_preserved_information", IntType, 0),
            ("max_cells", IntType, 0),
            ("max_total_parts", IntType, 0),
            ("max_part_numbers", DictType(StringLikeType, IntType), None),
            ("results_prefix", StringLikeType, ""),
        ],
        [
            "coclustering_file_path",
            "simplified_coclustering_file_path",
            "results_dir",
        ],
        # fmt: off
        """
        // Simplify coclustering settings
        LearningTools.PostProcessCoclustering
        SelectInputCoclustering
        InputCoclusteringFileName __coclustering_file_path__
        OK
        PostProcessingSpec.MaxPreservedInformation __max_preserved_information__
        PostProcessingSpec.MaxCellNumber __max_cells__
        PostProcessingSpec.MaxTotalPartNumber __max_total_parts__
        __DICT__
        __max_part_numbers__
        PostProcessingSpec.PostProcessedAttributes.List.Key
        PostProcessingSpec.PostProcessedAttributes.MaxPartNumber
        __END_DICT__

        // Output settings
        AnalysisResults.ResultFilesDirectory __results_dir__
        AnalysisResults.ResultFilesPrefix __results_prefix__
        AnalysisResults.PostProcessedCoclusteringFileName __simplified_coclustering_file_path__

        // Simplify Coclustering
        PostProcessCoclustering
        Exit
        """,
        # fmt: on
    ),
]
